// _eEspruinoJsShim.js - adding Espruino js built-in things to browser js context

// 20191117 - (c) allObject

// Used in Espruino cross development environment in Browser - extend as needed.
//
// Includes so far:
// -- getTime(), system time in SECONDS
// -- setWatch() / clearWatch()
// -- Pin.high() / Pin.low() (PIN/BTN... if watched,...)
// -- digitalWrite()
// -- _eConsole singleton w/ .log(), .warn(), .error()
// -- require() 
// -- Modules - modules cache supporting require()


// === Support for emulation console (eConsole) logging (by default to standard console)
var _eConsole =
{ console: console // for opt. override w/ custom console (and keep _eLogNode null).
, consoleOn: true  // log to browser console as well
, _eLogNode: null  // if a (DOM) node (!= null), prepend as line to DOM node's body
, log:   function() { if (this._eLogNode) { this._eLog("I",arguments);
           } else { this.console.log.apply(  this.console, arguments); } }
, warn:  function() { if (this._eLogNode) { this._eLog("W",arguments);
           } else { this.console.warn.apply( this.console, arguments); } }
, error: function() { if (this._eLogNode) { this._eLog("E",arguments);
           } else { this.console.error.apply(this.console, arguments); } }
, _eLog: function(typ,args) { // log message type "I|W|E|orCustom", args / array
    var ss = [], sc = [], _eTs = ("000000" + (new Date().getTime() % 100000)).substr(-6)
      , cs = "|......................." // 24
      , cf = "|......................................." // 40
      , s, sfs, sfc;
    for (var i=0; i<args.length; i++) {
      if ((typeof (s = args[i]) == "string") && !(s.indexOf(cs))) {
        // console.log(s);
        sfs
        = (s.substr(-5) == "|btn|") ? cf.substr(0,20)
        : (s.substr(-5) == "|chk|") ? cf.substr(0,20)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,22)
        : (s.substr(-5) == "|sli|") ? cf.substr(0,20)
        : cf.substr(0,24);
        ss.push(sfs);
        sfc
        = (s.substr(-5) == "|btn|") ? cf.substr(0,24)
        : (s.substr(-5) == "|chk|") ? cf.substr(0,24)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,26)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,24)
        : cf.substr(0,24);
        sc.push(sfc);
      } else {
        ss.push(s);
        sc.push(s);
      }
    }
    if (this.consoleOn) { sc.splice(0,0,[_eTs]); console.log.apply(console,sc); }
    this._eLogNode.innerHTML = _eTs + ": " + typ + ": " + ss.join(", ")
                             + "<br>" + this._eLogNode.innerHTML; }
, _eRegisterLogNode: function(id) { this._eLogNode = document.getElementById(id);
    this.log( "_eEmulationPath:"   + _eEmulationPath
            , "_eModulesPath:"     + _eModulesPath
            , "_eProjectsPath:"    + _eProjectsPath
            , "_eMinifiedDefault:" + _eMinifiedDefault
            ); }
};

// some dynamic hatml support for X dev html files
var _efs = []; // html fragments // re-init before use
function _ep(s) { _efs.push(s); } // add s to html fragments
function _ew() { document.writeln(_efs.join("")); _efs = []; } // write frags to html

// ===== Ports / Pins and Appliances

var Pin = function(id) { this.id = id; this.oIds = [];
                         this.btn = false;
                         this.ov = 0;
                         this.iv = 0;
                         this.time = 0;
                         this.lastTime = 0;
                         this.risingWatcher = null;
                         this.fallingWatcher = null;
                         this.ad = "";
                       };
Pin.prototype.other    = function(n,isBtn) { this.oIds.push(n); this.btn=!!isBtn; return this; };
Pin.prototype.toString = function() { return (this.id + ":" + this.v); };
Pin.prototype.isBtn    = function() { return (this.btn); } 
Pin.prototype.set = function() {
  if (this.iv === 1) {
    _eConsole.log(this.id + ".set() NO change " + ((this.ad) ? " - " + ad : "")); 
  } else {
    _eConsole.log(this.id + ".set() A change! " + ((this.ad) ? " - " + ad : "")); 
    this.lastTime = this.time; this.time = getTime();
    this.ov = 1;
    if (this.risingWatcher) this.risingWatcher.evt(true); } }
Pin.prototype.reset = function() {
  if (this.iv === 0) {
    _eConsole.log(this.id + ".reset() NO change " + ((this.ad) ? " - " + ad : "")); 
  } else {
    _eConsole.log(this.id + ".reset() A change! " + ((this.ad) ? " - " + ad : "")); 
    this.lastTime = this.time; this.time = getTime();
    this.ov = 0;
    if (this.fallingWatcher) this.fallingWatcher.evt(false); } };
Pin.prototype.high = function() {
  if (this.iv === 1) {
    _eConsole.log(this.id + ".high() NO change " + ((this.ad) ? " - " + ad : "")); 
  } else {
    _eConsole.log(this.id + ".high() A change! " + ((this.ad) ? " - " + ad : "")); 
    this.lastTime = this.time; this.time = getTime();
    this.iv = 1;
    if (this.risingWatcher) this.risingWatcher.evt(true); } }
Pin.prototype.low = function() {
  if (this.iv === 0) {
    _eConsole.log(this.id + ".low() NO change " + ((this.ad) ? " - " + ad : "")); 
  } else {
    _eConsole.log(this.id + ".low() A change! " + ((this.ad) ? " - " + ad : "")); 
    this.lastTime = this.time; this.time = getTime();
    this.iv = 0;
    if (this.fallingWatcher) this.fallingWatcher.evt(false); } };

var SPI = function(id) { this.id = id; this.ids = ["id"]; }
SPI.prototype.toString = function()  { return this.id; };
SPI.prototype.setup    = function()  { log(this.id + ".setup()"); };


// ===== functions

function require(moduleName) {
  _eConsole.log('require("' + moduleName + '");');
  return Modules._eCached(moduleName);
}

function digitalWrite(pin,value) {
  if (value) pin.set(); else pin.reset();
}

function getTime() {  // returns system time in seconds as floating number
  return (new Date().getTime()) / 1000;
}

// ===== setWatch() / clearWatch() - including supporting -e-'class' and singleton

function setWatch(f,pin,opts) { _eWatches.setWatch(f,pin,opts); }
function clearWatch(id) { _eWatches.clearWatch(id); }

var _eWatch = function(wId,f,pin,opts) { // Watch class
  this.wId = wId;
  this.f = f;
  this.pin = pin;
  this.opts = opts;
  if ((opts.edge == "rising" ) || (opts.edge == "both")) pin.risingWatcher  = this;
  if ((opts.edge == "falling") || (opts.edge == "both")) pin.fallingWatcher = this;
};
_eWatch.prototype.evt = function(edgeEnd) {
  if (    ("both" == this.opts.edge)
       || (   edgeEnd && (this.opts.edge == "rising" ))
       || ( ! edgeEnd && (this.opts.edge == "falling"))
     ) {
    _eConsole.log(this.pin.id+":"
       +" "+this.opts.edge
       +((this.opts.edge === "both")
         ? " ("+((edgeEnd)?"rising":"falling")+")"
         : "")
       +" watch event happened");
    var e =
      { state: edgeEnd
      , lastTime: this.pin.lastTime
      , time: this.pin.time
      };
    setTimeout(this.f,1,e);
  }
}
_eWatch.prototype.unwatch = function() {
  if ((this.opts.edge == "rising" ) || (opts.edge == "both")) this.pin.risingWatcher  = null;
  if ((this.opts.edge == "falling") || (opts.edge == "both")) this.pin.fallingWatcher = null;
}

var _eWatches = // singleton holding on to watches
{ lastId: 0
, watches: {}
, setWatch: function(f,pin,opts) {
    var wId;
    opts = opts || { repeat:true, edge:"both", debounce:10 }
    opts.repeat = opts.repeat || false;
    opts.debounce = 
      (opts.debounce !== undefined)
        ? opts.debounce
        : (pin.isBtn())
          ? 10
          : 0;
    opts.edge = opts.edge || "both";
    if (    (pin.risingWatcher  && ((opts.edge == "rising" ) || (opts.edge === "both")))
         || (pin.fallingWatcher && ((opts.edge == "falling") || (opts.edge === "both")))
       ) {
      _eConsole.err("Unable to set "+opts.edge+" watch on "+pin+";"
         +" watch on "
         +" ("+((pin.risingWatcher && pin.fallingWatcher)?"both"
              :((pin.risingWatcher)?"rising":"falling"))
         +" is already taken.");
      } else {
        this.watches["w"+(wId=++this.lastId)]
          = new _eWatch(this.lastId,f,pin,opts);
      }
      return wId;
    }
, clearWatch: function(id) {
    var w = this.watches["w"+id];
    if (w) {
      w.unwatch();
      delete this.watches[wId];
    } else {
      _eConsole.err("Unknown watch id " + id);
    }
  }
};


// ==== Modules cache
var Modules =
{ modules: {}
, removeCached: function(id) {
    if (this.modules[id]) {
      delete this.modules[id];
    } else {
      _eConsole.warn("Modules.removeCached(): "
          + "module " + id + " not found or already removed from cache");
    }
  }
};

// ==== Support for emulated Modules cache
Modules._eCache = function(moduleName, module) {
    _eConsole.log('Modules._eCache("' + moduleName + '", ...);');
    this.modules[moduleName] = module;
  };
Modules._eCached = function(moduleName) {
    var module = this.modules[moduleName];
    if (!module) {
      if (!this.modules.hasOwnProperty(moduleName)) {
        this.modules[moduleName] = undefined;
      }
      _eConsole.error("ERROR: module with name " + moduleName + " not found.");
    }
    return module;
  };
// stores args for emulated modules at _emLoad()
Modules._eeArgs = {};
Modules._eeRegisterArgs = function(args) {
    var ss = []; for (var i=0; i<args.length; i++) { ss.push(args[i]); }
    _eConsole.log("Modules._eeRegisterArgs() - "
        + args.length + ": [ " +ss.join(", ") + " ]");
    this._eeArgs[args[0]] = args;
  };
// support pre load in html (assumes _eExmpl to be global indicating example name);
// mExpl/s is single or array of module (ui element) example as selector for global
// _emExmpl to catch on and add single or array of module/s, and - if pExample is
// present - update global _epExmpl (for cases where on project example show case
// more than one ui element or shares the same project example, because it needs
// this or these other ui elements.
Modules._esToLoad = [];
Modules._eToLoad = function(mExmplOrmExmpls,moduleOrmodules,pExmpl) {
    var ms = this._esToLoad, x; // skip modules that are already in ._esToLoad[]
    (("string" == typeof mExmplOrmExmpls)?[mExmplOrmExmpls]:mExmplOrmExmpls
    ).forEach(function(ex){
      if (ex===_emExmpl) { // <----------- global _emExmpl
        (("string" == typeof moduleOrmodules)?[moduleOrmodules]:moduleOrmodules
        ).forEach(function(m){ x=ms.length;
          while (--x >= 0) if (m==ms[x]) x = -2;
          if (x===-1) ms.push(m);
        },this);
        // overwrite pre set _epExmpl when pExmpl
        // BUT ONLY when _emEmpl not 1st in mExmplOrmExmpls
        // (enables reuse of project file for individual mExmplOrmExmpls
        if (    (_emExmpl !== mExmplOrmExmpls[0]) 
             && pExmpl ) {
          log("_epExmpl adjusting to " + pExmpl + "(from" + _epExmpl + "; example=" + _emExmpl + ")");
          _epExmpl = pExmpl; // <----------- global _epExmpl
        }
      }
    },this);
  };
Modules._eLoad = function() {
    this._esToLoad.forEach(function(m){ _emLoad(m); });
  };


// === Support for module load emulation (from modules folder);
// is part of emulated upload to board with the following difference:
// upload to board does parse recursively / 'nested' all code with
// regex before uploading to have all dependencies loaded. In emulation,
// This happens on html page load and is hardcoded / 'manually' *BEFORE*
// Espruino code by dynamically pulling module as src in html script tag.

var exports;
function _emLoad(moduleName) {
  _eConsole.log('_emLoad("' + moduleName + '"...);');
  document.writeln(['<scri','pt>'
      ,'exports={};'
      ,'</scri','pt>'].join(""));
  document.writeln(['<scri','pt src="'
      ,_eModulesPath,moduleName,
      ,((_eMinifiedDefault) ? ".min" : ""),'.js">'
      ,'</scri','pt>'].join(""));
  document.writeln(['<scri','pt>'
      ,'Modules._eCache("',moduleName,'", exports);'
      ,'</scri','pt>'].join(""));
}


// === Support for project load emulation (from projects folder);
// This allows to pull in / embed the Espruino upload code as is
function _epLoad(projectName) {
  _eConsole.log('_epLoad("' + projectName + '"...);');
  document.writeln(['<scri','pt src="'
      ,_eProjectsPath,projectName,'.js">'
      ,'</scri','pt>'].join("")); }

// board flags - respective _eIsXyz is set in _eEspruinoBoardShimXyz.js
var _eIsOriginal
  = _eIsPico
  = _eIsWifi
  = _eIsPixl
  = _eIsBangle
  = false;
