// uiExampleAllPixl.js - example w/ all ui elements on Pixl.js

// 201911108 - allObjects - vars: ~ / ~

// ----- setup lon and log() (considering emulating _eConsole)
var lon = true // logging is on / off
  , log = function() { if (lon) { var lgr;(lgr=(typeof _eConsole
       =="object")?_eConsole:console).log.apply(lgr,arguments);}};

// ----- pull in ui base and ui element(s)
var ui = require("ui")     // basic module (ncommented)
    .adx(require("uiE3P")) // ui extension for 3 button controlled Pixl
    .adx(require("uiBtn")) // ui plain button element
    .adx(require("uiChk")) // ui check box element
    .adx(require("uiRad")) // ui radio button element
    .adx(require("uiSli")) // ui slider element
    .adx(require("Font6x8"),-2,"Font6x8") // pixl suitable bitmap font (fs=-1)
    ;

// ----- define callback(s): preferred (untouch) / generic (all) touch event(s)

var cb = function(id, v, _, e) { // preferred - untouch event only
           log("id:",id,"v:",v,((_.t)?"":"un")+"touch focused",_.x,_.y,_.t); }
  , ca = ( (typeof ca === "undefined") // generic - all touch events
         ? function(id, v, _, e) { var s=(_.f|65536).toString(2); // log flags
             log("c:",e[1],"c:",e[1],"id:",id,"v:",v,"f:"
                ,[s.substr(5,4),s.substr(9,4),s.substr(13,4)].join(" "),_.f); }
         : ca )
  , backlite = function(id, v, e) { // turn backlight on / off w/ label change
         digitalWrite(LED,v); // e[11][4]=(v)?"off":"on"; // set LED and label
         return false; } // return true/y in callback redraws ui elt & label(s)
  ;


// ----- create UI (executing definition in function requires less memory on upload)

var s1 // hold on to "s1" slider ui element for direct access app later on.
  , cui = function() { var u = undefined; // invoke as 1st in onInit()

// ---------- Button examples ----------
//....0  1     2       3   4   5   6   7   8  9       10  11                         12
//.flgs  clazz id      x   y   w   h  bc  fc  valObj  cb,  l (label array obj)
//       btn                ->x2->y2                        fs tc    x  y label txt  ca
ui.c( 3,"btn","b1"  ,  0,  0, 30, 21,  0,  0, "B_1",  cb,  [-1, 7,   7, 6, "Reg" ]);
ui.c( 3,"btn","b2"  , 29,  0, 41, 21,  0,  7, {v:2},  ca , [-1, 0,   7, 6, "BoYbg"],ca);
ui.c( 3,"btn","b3"  , 70,  0, 34, 14,  0,  7, 3    ,  cb , [ 0, 0,   5, 4, "Magnum"]);
ui.c( 3,"btn","b4"  ,103,  0, 26, 12,  0,  0, 4    ,  cb , [ 0, 7,   5, 3, "Tiny"]);

// ---------- Checkbox examples ----------
//....0  1      2      3   4   5   6   7   8  9       10  11                         12
//.flgs  clazz id      x   y   w   s  bc  fc  valObj  cb,  l (label array obj)
//       chk                ->x2+>y2~------->[s,vOjb]       fs tc    x  y label txt
ui.c( 3,"chk","c2"  , 70, 16, 13,  0,  0,  0, "X",    u ,  [-1, 0,   1, 3, "Up"   ], ca);
ui.c( 3,"chk","c1"  , 95, 16, 13,  1,  0,  7,true,backlite,[ 0, 0,   1, 8, "light"]);

// ---------- Radio button examples ----------
//....0  1     2       3   4   5   6   7   8  9       10  11                         12
//.flgs  clazz id.grp  x   y   w   s  bc  fc   valObj cb,  l (label array obj)
//       rad                ->x2+>y2~------->[s,vOjb][g,cb,]fs tc    x  y  label txt ca
ui.c( 3,"rad","r1.a",  0, 23, 23,  0,  0,  0, "L",    cb,  [ 0, 0,   2, 9, "Large"]);
ui.c( 3,"rad","r1.b", 45, 26, 17,  1,  0,  0, "M",    0 ,  [ 0, 0,   2, 7, "Med"  ]);
ui.c( 3,"rad","r1.c", 77, 31, 11,  1,  0,  0, "S",    0 ,  [ 0, 0,   2, 3, "Small"]);

// ---------- Slider examples ----------
s1 = // hold on to "s1" slider ui element for direct access app later on.
//....0  1     2       3   4   5   6   7   8  9       10  11                         12
//.flgs  clazz id      x   y   w   h  bc fcs  value   cb, ls (arr of labels ([[...],..]))
//       sli                ->x2->y2                    [0] fs tc    x  y  txt // formatter
ui.c( 3,"sli","s1"  , 12, 46, 82, 16,  0,[0,7],  33,  ca, [[ 0, u,  30, 7,   u
                                                           ,function(v,_,l){ // val,ui,labl
           var s="", f=_.f; return [(!(f===0 || f&184)) ? s : ((v<1)?"0":"") // 128+32+16+8
                       +(s=_.mr(v*10)+"%").substr(0,s.length-2)+"."+s.substr(-2)]; } // val
//                         _.mr()=Math.round                min max si  o a flgs
                                                           ,  0,100,[3],0,1, 168
//                                                          left top right bot margins
                                                           ,   0,  3]
//                                                     [>0] fs tc    x  y  txt   fmt
                                                          ,[ 0, 0, -12, 7, "%:0"] // frnt
                                                          ,[ 0, 0,  84, 7, "100"] // back
                                                          ]);
//....0  1     2       3   4   5   6   7   8  9       10  11                         12
//.flgs  clazz id      x   y   w   h  bc fcs  value   cb, ls (arr of labels ([[...],..]))
ui.c( 3,"sli","s2"  ,118, 31, 10, 33,  0,[0,7],   6,function(id,v,_,e){ e[9]=_.mr(v); } // i
//       sli                ->x2->y2                    [0] fs tc    x  y  txt // formatter
                                                         ,[[ 0, 0, -12,14, ":-"
                                                           ,function(v,_,l){ // val,ui,labl
          var s="",f=_.f; return [(!(f===0||f&184))?s:((v>=0)?" ":"")+v+l[4] // 128+32+16+8
                                 ,[7,0,0, 8, 6]]; } // clear bg rect w/ clr, x/y offset+ext
//                                                          min max si  o a flgs
                                                           , -9,  9,[1],1,0, 168
//                                                          left top right bot margins
                                                           ,   2]
//                                                     [>0] fs tc    x  y  txt   fmt
                                                          ,[ 0, 0,  -4, 2,  "9"] // top
                                                          ,[ 0, 0,  -8,26, "-9"] // bot
                                                          ]);

};

//---------- extra drawings examples ----------
var xDs = function(ui) { // e (2nd arg of no use on draw all)
  ui.dl([0,0,108, 16, "back"]);
  var d = ui.dsp     // display - Espruino Graphics object
    , e = s1         // =ui.e("s1") // get slider ui element w/ id "s1" 
    , x = e[3]       // get e's left - x
    , m = e[11][0]   // to access margins 12,13,14/left,top,right
    , b = x+2+m[12]        // start of slider bar (border, value 0)
    , l = e[4]+2+m[13]     // get e's top-1=y-1=lower y of tick (1 pixel above  bar)
    , w = e[5]-b-2-1-m[14] // get e's slider bar width (right border (-1?), val 100)
    , v = w/100  // pixels per 1% (100% = w pixels)
    , u = l-2    // upper y of tick, use all 4 margin pixels betw. bar & sel/focus line
    , t // tick pos
    ;
  ui.clr(e[7]); // same as slider bar's border color
  for (var i=0;i<=100;i+=10) {
    t = ui.mc(b + i * v) + 0.5; // weird +0.5, only in emulation / canvas
    d.drawLine(t,l,t,u-((i % 50)?0:2)); // short (2px) and long (4px) ticks
  }
};

// ----- Pixl specific
var dsp = g; // ...to keep var names to same in all examples

// ----- run UI
function onInit() {
  cui(); // create ui
  dsp.clear(); LED.set(); // display clear and turn back light on
  setTimeout(function(){
    // connect ui to dsp and display it
    ui.connect(dsp,BTN3,BTN4,BTN1)
      .dr(1)      // wipe screen w/ background color using draw rect
      .d(0,xDs)   // display all elements
      .di = true; // set display draw / changes to immediate
  },1000);
}
