/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- ui ------------------------- / 254 vars

*** modular and extensible ui framework.

to do: 
- describe how callback works (change ui elt and returning true redraws)
- describe/indicate optional values

The ```ui``` framework or base takes care of most of the plumbing and flow logic
that goes on between touch screen (or physical buttons) and display as input and
output components. ```ui``` is currently implemented as singleton. At runtime it
does not only hold on to its own code and data but also the code and data of the
on demand added and created ui elements.

Currently, plain buttons, check box, radio button, slider, input field - and with
last for course - a keyboard are available. Extensions are easily to create and
integrate based on the architecture and the available documentation.

All ui elements handled by ui framework have roughly 12 common properties in same
structure: flags, clazz, x and y position, with and height, border and fill colors,
a value object, a callback and some label object(s), of which only the 9 1st are
really mandatory. Same values in same structure lead to implementation of the ui
elements as lightweight, lean array objects in order to save memory.

*** Basic Application of ui base and - as example - uiBtn (Button) element

--- get ui base and ui elements as needed, such as uiBtn, uiSli(der), ui...
Get ui base and ui elements and build ui from ui definitions in level 0 to save vars,
where it is executed on Espruino immediately on upload and this ui definition source
code is not stored. save() will save the built ui and start from it onInit() on
power up.

```
var ui = require("ui")       // getting ui base and control code (default 8 colors)
      .add(require("uiBtn")) // add/mixin btn ui elt support (and remove from cache)
      ;
```

--- define UI and add elements on code upload - but do not display it yet 
When UI is built in level 0, build happens at upload to the board and source code
of UI definition is not consuming any runtime memory in order to save variables.

```
//   0  1     2       3  4     5    6    7   8  9        10  11 
// flgs clazz id      x  y     w    h   bc  fc  valObj   cb,  l (label array obj)
//      btn                 ->x2 ->y2                         fv tc    x  y  label text
ui.c(3,"btn","b1"  ,  5, 40,  65,  35,  4,   4, "B_1",   cb,  [15, 7,  13, 9, "RED"  ]);
ui.c(3,"btn","b2"  , 70, 40,  65,  35,  5,   6, {v:1},   cb,  [15, 0,   9, 9, "Y-w-B"]);
// bc/fc/tc = border/fill/text colors; cb=callback(id, v, ui, e, t). id is button id,
// such as "b01"; v, ui, e, and t provide button, ui, and touch event runtime data.
// colors are 1-bit color dept coded b&rgb (0=black, 7 is white, 4 is red, etc) 
```

--- devices stuff - done on upload - once 

```
var dsp,   dspMod   = require("ILI9341");    // display
var touch, touchMod = require("TouchScrRD"); // touch screen w/o controller
```

// --- device stuff in onInit(){...} - done on power up - every time (and on save())        s

```
function onInit() {
  SPI1.setup({sck:B3, miso:B4, mosi:B5, baud: 1000000});
  dsp = dspMod.connect(SPI1, B6, B8, B7, function() {
    dsp.clear(); B2.set(); // clear display and switch backlight on
    ui.connect(dsp)        // connect ui to dsp (Espruino Graphics object)
      .w(0,0,240,320)      // wipe (clear/fill) screen with screen background color
      .d()                 // display all ui elements
      .di=true;            // set to instant display
    touch = touchMod.connect(C0,C1,C2,C3,function(x,y){ ui.evt(x,y); }); // ...to ui
    touchr.listen(true);   // not applicable for ADS7843 module
  });
} // /onInit()
```

ui provides also convenience functions/methods to use in application to deal with
ui elements as well as plain graphics. Due to ui's extensibility, you can add easily
your own ui element or plain function extensions as desired (with .add(module, keep);).

- .iib(x,y,e) returns true when x and y lay in bounding box of ui element e
- .foc(e) set focus to element e (and take it away from element that had it so far)
- .blr(e) blur/unfocus element e (if .blr(), element currently in focus is blurred)
- .w(x,y,w,h,c) wipe rectangular area at x/y and w/h of display with color c
- .c(arguments) generic, single point error for creating ui elements. Uses first
     arguments[1] as class or type name and concatenates it with "C" for specific
     create entry. To create a custom ui element and integrate it (with
     .add(require("customUiElementModule")) is discussed in a separate publication.
- .d(e) redraw element e, and when just .d(), redraw the whole ui (all ui elements). In
     general, a ui is built on upload without displaying it to save variables and 
     displayed on demand or in onInit() to re-display on every re-power up or reset.
- .clr(c,i) set color according this.clrs{0](c,i) (customizable) function. Default is
     set for bit coded value for 1-bit (rgb) color depth and provides 8 colors:
     b&000=0=black, b&111=7=white, b&100=4=red, etc. this.clrs[0]() can be customized
     to support any color depth - coded, table looked up, or literal. For details see
     Colors section.
- .fnt(fv) set fontVector (size) ...currently w/ no frills, as opposed to .clr().
     More font options will be implemented at a later time. 
- .ld(x,y,l(abelInfoArray)) label draw function; label info array:
     l[0] : fontVector (size)
     l[1] : color (see Colors section
     l[2] : xOffset to x - left of bounding box of label (.drawString(...))
     l[3] : yOffset to y - top  of bounding box of label (.drawString(...))
     l[4] : label text (used in .drawString(...))
     Note: the reason for x and xOffset - and y and yOffset - is because .ld() is
     used to in conjunction with an ui element that has a bounding box that provides
     x and y - usually right-top corner - and and xOffset and yOffset are used to
     place the label relative to the ui elements supplied corner coordinates.
- .add(module,keep) extend ui module with ui element and other extensions to support
     modularity. After module code is merged into ui base code, module is removed from
     cache to save variables. For that purpose, the module has to know its name in mn
     property. For example, for a module onOff.js switching backlight on/off on pin B2,
     exports = { mn:"onOff", on: function(on) { digitalWrite(B2, on || on===undefined); }
     defines the module code, ui.add(require("onOff")) adds it to be used as ui.on();.
     Without building a file, modules can also be put into cache directly on upload in 
     level 0 (and served with require() with module name provided as variable to escape 
     regular module detection and pre-upload from modules sources):
     Modules.addCached(mName,'{ mn:"onOff",on: function(on) { digitalWrite(B2,'
       +' on || on===undefined); }');, and added with short ui.add("onOff");

UI (touch) event object:
    The touch event object ```t``` and other objects, such as the ```ui```/```_``` and
the ui element ```e``` themselves, are passed around for easy access in the ui element
implementation as well as in the application. The touch event object has this structure:

```
t = // (touch) event
   { x: x // current or last touched x coordinate
   , y: y // current or last touched y coordinate
   , t: t // touching: (truey), not touching (anymore): falsy
   }
```
 
Colors:
    Colors can be specified in three ways: bit coded, index into (r,g,b) triplet, and 
literal (r,g,b) triplet. For (r,g,b) specification, Espruino Grapics modules accepts for
each r,g,b the normalized value 0..1. Under the hood and with help of the module for the
presenting display, this normalized color specification is transformed - most of the time
into a bit coded color again, but coded in different, more complicated way, such as
not equal number of bits for each base color r, g, b, for example (hardware display 
specific), 5, 6, 5 bits, and sometimes different r, g, b sequence.

Also the sequence sequence of the r, g, b, information c 
For each ways, color depth matters, in order to pass on 
normalized color specification to Espruino's Graphic object. Therefore, ui 
color depth implied
ui module already includes a default color converter from bit coded rgb value to
Espruino's normalized [r,g,b] triplet which makes bit coded colors independent from
the display's version of bit coding. ui's default color converter supports 1-bit color
depth coded colors - 2^1^3 = 8 colors.
  
 */
exports = // ui base / 'DOM/ui e(lement) data & code holder, singleton, for mixins)
{ dsp: null // dsp (Espruino Graphics object)
, mn: "ui"  // reserved, used temporary
, bc: 0     // (display) background color r*g*b* bit coded or if <0 looked up [R,G,B]
, tc: 7     // touch color r*g*b* bit coded or if <0 looked up [R,G,B]
, di: false // display instantly on create/change (not defer until after save())
, es: []    // ui elements
, ef: null  // (ui) element (in) focus
, lf: null  // (ui element) last (in) focus
, it: false // (display) is touched
, td: false // is touch down (event)
, tt: 0     // touch down time
, lx: 0     // last x (touched)
, ly: 0     // last y (touched)
, dx: -1    // (last) x (touch) down
, dy: -1    // (last) y (touch) down
, te: {t:0} // (last) touch event
, clrs:     // default - bit coded color of 1 bit color-depth - 2^1^3(rgb) = 8 colors
    [function(c,i){ var v=(i)?c^7:c; return [v>>2&1,v>>1&1,v&1]; }.bind(exports)
    ]
, evt: function(x,y,_) { // common handling of touch / untouch event
    _=_||this; var a=_.es, i=-1, m=a.length, e=_.ef, t={t:undefined!==x}; _.te=t;
    if (_.td = !_.it && t.t) { _.dx=x; _.dy=y; _.tt=getTime(); } // touch down
    if (_.it=t.t) { _.lx=t.x=x; _.ly=t.y=y; // touching
      if (e) { if (_.iib(e,x,y)) { return _[e[1]](_,e,t); } // xy event/move in focus elt
        _.blr(e) } // just moved out
      while (++i<m) { if (((e=a[i])[0]&2) && _.iib(e,x,y)) { // loop through active elts
        if (!(e[0]&4) && (_.td || (t.t && (e===_.lf)))) { _.foc(e); } // adj focus
        _[e[1]](_,e,t); return; } } if (_.td) { _.lf=null; } // handle hit / miss
    } else { if (e) { t.x=_.lx; t.y=_.ly; // untouch
      if (!_[e[1]](_,e,t)) { _.blr(e); } _.lf=null; _.dx=_.dy=-1; } } }
, iib: function(e,x,y) { // (if x y) is in (element bounding) box return true, else false)
    return ((x>=e[3])&&(y>=e[4])&&(x<=e[5])&&(y<=e[6])); }
, foc: function(e,_) { // focus elt (unfocus elt in focus); visual: draw rectangle in...
    _=_||this; if (e!==_.ef) { if (_.ef) { _.blr(_.ef); } // ...touch color around elt
    _.clr(_.tc).dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]|=4; _.ef=_.lf=e; }
    return _; }
, blr: function(e,_) { // blur elt (all if e falsy); visual: draw rectangle in display...
    _=_||this; _.clr(_.bc); (e?[e]:_.es).forEach(function(e){ // ...background clr around 
        if (e[0]&4) { _.dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]&=3; } }); _.ef=null;
    return _; }
, rdc: function(e) { // register and return ui element and draw/display it conditionally
    e[0]&=7; this.es.push(e); if (this.di) { this.d(e); }
    return e; }
, w: function(x,y,x2,y2,c) { // wipe rectangle (fill w/ c(olor) or display backgrnd color)
    this.clr((c===undefined)?this.bc:c); this.dsp.fillRect(x,y,x2,y2); return this; }
, c: function() { // c(reate) ui element like DOM.createElement(), where arguments[1]...
    // ...defines ui element clazz and points to ui element specific constructor method
    var a=arguments, c=this[a[1]+"C"]; if (c) {
    return c.apply(this,a); } }  // undefined ui classes returns undefined as ui element
, d: function(e) { // display / redraw ui element e, redraw all if e falsy
    ((e)?[e]:this.es).forEach(function(e){ this[e[1]+"D"].apply(this,e); },this);
    return this; }
, clr: function(c,i) { // set color according this.clrs[0](c,i) (customizable) function...
    this.dsp.setColor.apply(this.dsp,this.clrs[0](c,i)); // ...w/ color and i-nverse flag
    return this; }
, fnt: function(s) { // set displays's font vector (font size)
    this.dsp.setFontVector(s);
    return this; }
, ld: function(x,y,l) { // draw label (x, y, label[fv, tc, x, y, txt])
    this.fnt(l[0]).clr(l[1]).dsp.drawString(l[4],x+l[2],y+l[3]); }
, add: function(emon, k) { // add (mixin) non-overwriting properties of ui element by... 
    // eltModuleOrName for modularization and extensibility. Remove module from cache...
    var p, // ...when not k(eep) by element module name - if defined - to save space
    m = ("string" === typeof emon) ? require(emon) : emon; for (p in m) { if ((p!=="mn")
    &&!this[p]) { this[p] = m[p]; } } // merge / mixin properties 'new' (vars, functions)
    if (!k) { Modules.removeCached(m.mn); } // delete module from cache when not k(eep)
    return this; }
, connect: function(dsp) { // connects w/ initialized Espruino Grapics obj (after save()) 
    this.dsp=dsp;
    return this; }
};
