/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- ui ------------------------- / 258 vars (on empty of 27)

*** Basic Application of ui base and uiBtn (Button) element

--- get ui base and ui elements as needed, such as uiBtn, uiSli(der), ui...
Get ui base and ui elements and build ui from ui definitions in level 0 to save vars,
where it is executed on Espruino immediately on upload and this ui definition source
code is not stored. save() will save the built ui and start from it onInit() on
power up.

```
var ui = require("ui")        // getting ui base and control code (default 8 colors)
      .add(require("uiBtn")); // add/mixin btn ui elt support (and remove from cache)
```

--- define UI and add elements on code upload - but do not display it yet 
When UI is built in level 0, build happens at upload to the board and source code
of UI definition is not consuming any runtime memory in order to save variables.

```
//   0  1     2       3  4     5    6    7   8  9       10  11 
// flgs clazz id      x  y     w    h   bc  fc  valObj  cb,  l (label array obj)
//      btn                 ->x2 ->y2                         fv tc    x  y  label text
ui.c(3,"btn","b1"  ,  5, 40,  65,  35,  4,   4, "B_1",  cb,  [15, 7,  13, 9, "RED"  ]);
ui.c(3,"btn","b2"  , 70, 40,  65,  35,  5,   6, {v:1},  cb,  [15, 0,   9, 9, "Y-w-B"]);
// bc/fc/tc = border/fill/text colors; cb=callback(id, v, ui, e, t). id is button id,
// such as "b01"; v, ui, e, and t provide button, ui, and touch event runtime data.
// colors are 1-bit color dept coded b&rgb (0=black, 7 is white, 4 is red, etc) 
```
 
--- devices stuff - done on upload - once 

```
var dsp,   dspMod   = require("ILI9341");    // display
var touch, touchMod = require("TouchScrRD"); // touch screen w/o controller
```

// --- device stuff in onInit(){...} - done on power up - every time (and on save())        s

```
function onInit() {
  SPI1.setup({sck:B3, miso:B4, mosi:B5, baud: 1000000});
  dsp = dspMod.connect(SPI1, B6, B8, B7, function() {
    dsp.clear(); B2.set(); // clear display and switch backlight on
    ui.connect(dsp)        // connect ui to dsp (Espruino Graphics object)
      .w(0,0,240,320)      // wipe (clear/fill) screen with screen background color
      .d()                 // display all ui elements
      .di=true;            // set to instant display
    touch = touchMod.connect(C0,C1,C2,C3,function(x,y){ ui.evt(x,y); }); // ...to ui
    touchr.listen(true);   // not applicable for ADS7843 module
  });
} // /onInit()
```

ui provides also convenience functions/methods to use in application to deal with
ui elements as well as plain graphics. Due to ui's extensibility, you can add easily
your own ui element or plain function extensions as desired (with .add(module, keep);).
- .iib(x,y,e) returns true when x and y lay in bounding box of ui element e
- .foc(e) set focus to element e (and take it away from element that had it so far)
- .blr(e) blur/unfocus element e (if .blr(), element currently in focus is blurred)
- .w(x,y,w,h,c) wipe rectangular area at x/y and w/h of display with color c
- .c(arguments) generic, single point error for creating ui elements. Uses first
     arguments[1] as class or type name and concatenates it with "C" for specific
     create entry. To create a custom ui element and integrate it (with
     .add(require("customUiElementModule"),"customUiElementModule")) is discussed in a 
     separate publication.
- .d(e) redraw element e, and when just .d(), redraw the whole ui (all ui elements). In
     general, a ui is built on upload without displaying it to save variables and 
     displayed on demand or in onInit() to re-display on every re-power up or reset.
- .clr(c,i) set color according this.clrs{0](c,i) (customizable) function. Default is
     set for bit coded value for 1-bit (rgb) color depth and provides 8 colors:
     b&000=0=black, b&111=7=white, b&100=4=red, etc. this.clrs[0]() can be customized
     to support any color depth - coded, table looked up, or literal. For details see
     Colors section.
- .fnt(fv) set fontVector (size) ...currently w/ no frills, as opposed to .clr().
     More font options will be implemented at a later time. 
- .ld(x,y,l(abelInfoArray)) label draw function; label info array:
     l[0] : fontVector (size)
     l[1] : color (see Colors section
     l[2] : xOffset to x - left of bounding box of label (.drawString(...))
     l[3] : yOffset to y - top  of bounding box of label (.drawString(...))
     l[4] : label text (used in .drawString(...))
- .add(module,keep) extend ui module with ui element and other extensions to support
     modularity. After module code is merged into ui base code, module is removed from
     cache to save variables. For that purpose, the module has to know its name in mn
     property. For example, for a module onOff.js switching backlight on/off on pin B2,
     exports = { mn:"onOff", on: function(on) { digitalWrite(B2, on || on===undefined); }
     defines the module code, ui.add(require("onOff")) adds it to be used as ui.on();.
     Without building a file, modules can also be put into cache directly on upload in 
     level 0 (and served with require() with module name provided as variable to escape 
     regular module detection and pre-upload from modules sources):
     Modules.addCached(mName,'{ mn:"onOff",on: function(on) { digitalWrite(B2,'
       +' on || on===undefined); }');, and added with short ui.add("onOff");
       
Colors
Colors can be specified in three ways: bit coded, index into (r,g,b) triplet, and 
literal (r,g,b) triplet. For (r,g,b) specification, Espruino Grapics modules accepts for
each r,g,b the normalized value 0..1. Under the hood and with help of the module for the
presenting display, this normalized color specification is transformed - most of the time
into a bit coded color again, but coded in different, more complicated way, such as
not equal number of bits for each base color r, g, b, for example (hardware display 
specific), 5, 6, 5 bits, and sometimes different r, g, b sequence.

 Also the sequence sequence of the r, g, b, information c 
For each ways, color depth matters, in order to pass on 
normalized color specification to Espruino's Graphic object. Therefore, ui 
color depth implied
ui module already includes a default color converter from bit coded rgb value to
Espruino's normalized [r,g,b] triplet which makes bit coded colors independent from
the display's version of bit coding. ui's default color converter supports 1-bit color
depth coded colors - 2^1^3 = 8 colors.  
 */

exports = // ui base / 'DOM/ui e(lement) data & code holder, singleton, for mixins)
// t = // (touch) event
// { x: current or last touched x coordinate
// , y: current or last touched y coordinate
// , f: touch event flags:    untouch dragOver dragOut dragIn touchDown
//     0bxxx1 leave       &1     .       .        x       .         .
//     0bxx1x touch       &2     .       x        x       x         x
//     &bx1xx enter       &4     .       .        .       x         x
//     &b1xxx touchdown   &8     .       .        .       .         x       
// }                      //     0       2        5       6        14    
{ dsp: null // dsp (Espruino Graphics object)
, mn: "ui"  // reserved, used temporary
, bc: 0     // (display) background color r*g*b* bit coded or if <0 looked up [R,G,B]
, dc: 7     // focus color r*g*b* bit coded or if <0 looked up [R,G,B]
, tc: 6     // touch color r*g*b* bit coded or if <0 looked up [R,G,B]
, di: false // display instantly on create/change (not defer until after save())
, es: []    // ui elements
, ef: null  // (ui) element (in) focus
, et: undefined // (ui) element touched down
, it: false // (display) is touched
, td: false // is touch down (event)
, tt: 0     // touch down time
, lx: 0     // last x (touched)
, ly: 0     // last y (touched)
, dx: -1    // (last) x (touch) down
, dy: -1    // (last) y (touch) down
, clrs:     // default - bit coded color of 1 bit color-depth - 2^1^3(rgb) = 8 colors
    [function(c,i){ var v=(i)?c^7:c; return [v>>2&1,v>>1&1,v&1]; }.bind(exports)
    ]
, evt: function(x,y) { // common handling of touch down / drag / untouch event
    var _=this, e=_.ef, t={f:x===undefined)?0:2};
    if (f.f) { // touching
      _.td=!_.it; _.it=!0; _.lx=t.x=x; _.ly=t.y=y;
      if (e || _.et===e) { // focus is on touched down or some other focused
        if (_.iib(e,x,y)) { // xy inside it
          return _[e[1]](_,e,t); // acting on it while inside it
        } else { // xy just moved out of it
          t.t|=1; _.blrA(e,(_.et===e)?2:1,!_[e[1]](_,e,t))； // act on leaving it
        }
      }
      var a=_.es, i=-1, m=a.length; 
      while (++i<m) { if (((e=a[i])[0])&3) { if (_.iib(e,x,y)) {
        if (_.ef && (_.ef !== e)) { blrA(_.ef,!_[e[1]](_,e,t));
      
        if (_.td) { _.dx=x; _.dy=y; _.tt=getTime(); }; // touch down
        
        _.focB(e,(_.td)?2:1);
        return _[e[1]](_,e,t); } } }
      if (_.td) { _.lf=null; } // touch outside of any element
    } else { // is untouch
          _.blrA(e,e===_.et,_[e[1]](_,e,t)); // untouch
      if (e) { t.x=_.lx; t.y=_.ly; _[e[1]](_,e,t); _.ef=null; _.dx=-1; _.dy=-1; } } }
, iib: function(e,x,y) { // (if x y point) is in (element box return true, else false)
    return ((x>=e[3])&&(y>=e[4])&&(x<=e[5])&&(y<=e[6])); }
, fh: function(_,e,t,q) { // focus helper - returns e's focus (e[0]&4) as of touch event
    // and sets focus to e again when dragging into e and last touch down was on e)
    var f=(e[0]&12)>>2;
    if (!f  (_.td || (t.t && (e===_.lf)))) { _.foc(e); }
    return f; }
, focB: function(e,f) { // foc before invoking 
    var _=this, b=_.ef; 
    if (b && b!==e) { _.blrA(b,(b===_.et)?:); }
      
    return _; }
, foc: function(e,m) { // focus element (unfocus element currently in focus)
    // focus draws a rectangle around the ui element with touch focus color
    var _=this, fm=((m||1)&3)<<2, f=_.ef; if (f && (f !== e) ) { _.blr(f); }
    if ((e[0]&12)!==fm) { _.clr((fm&8)?_.tc:_.dc).dsp.drawRect(e[3],e[4],e[5],e[6]);
    e[0]|=fm; _.ef=_.lf=e; }
    log(e[2] + ".foc(): f:" + e[0]);
    return _; }
, blrA(e,f,p) { // blur post untouch (e=elt, d=touchDown, p=paint)
    if (f&2) { _.et=undefined; e[0]&=7; }; if (f&1) { _.ef=null; e[0]&=11; };
    if (p) { _.clr((d)?_.dc:_.bc).dsp.drawRect(e[3],e[4],e[5],e[6]); }
    return _; }
, blr: function(e,m) { // blur element e (any element if e falsy) to bit mask ((8|)4|0):
    // blur/unfocus draws rectangle around ui element in background or c(=focus) color
    var _=this, fm=((m||0)&3)<<2; _.clr((fm&4)?_.dc:_.bc); 
    ((e)?[e]:_.es).forEach(function(e){ if ((e[0]&12)>fm) { // change in focus cue
        _.dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]&(fm||3); } }); if (!fm) { _.ef=null; }
    log(e[2] + ".blr(): f:" + e[0]);
    return _; }
, rdc: function(e) { // register and return ui element and draw/display it conditionally
    e[0]&=15; this.es.push(e); if (this.di) { this.d(e); }
    console.log(e);
    return e; }
, w: function(x,y,x2,y2,c) { // wipe rectangle (fill w/ c(olor) or display background color)
    this.clr((c===undefined)?this.bc:c); this.dsp.fillRect(x,y,x2,y2); return this; }
, c: function() { // c(reate) ui element (like DOM.createElement(...) where arguments[1]...
    // ...defines ui element class and points to ui specific constructor method
    var _=this, a=arguments, c=_[a[1]+"C"]; if (c) {
    return c.apply(_,a); } }  // undefined ui classes return undefined
, d: function(e) { // display / redraw ui element e, do all if e falsy
    var _=this; ((e)?[e]:_.es).forEach(function(e){ _[e[1]+"D"].apply(_,e); });
    return _; }
, clr: function(c,i) { // set color according this.clrs{0](c,i) (customizable) function...
    var _=this; _.dsp.setColor.apply(_.dsp,_.clrs[0](c,i)); // ...and i-nverse flag
    return _; }
, fnt: function(s) { // set dsp's font vector (size)
    this.dsp.setFontVector(s);
    return this; }
, ld: function(x,y,l) { // draw label (x, y, label[fv, tc, x, y, txt])
    this.fnt(l[0]).clr(l[1]).dsp.drawString(l[4],x+l[2],y+l[3]); }
, add: function(emon, k) { // add (mixin) ui element properties by elementModuleOrName...
    // for modularization and extensibility. Remove module from cache when not to keep
    var _=this, p, // cache by element module name - if defined - to save space
    m = ("string" === typeof emon) ? require(emon) : emon;
    for (p in m) { _[p] = m[p]; } // merge / mixin modules properties (vars and functions)
    if (!k) { Modules.removeCached(_.mn); } delete _.mn; // delete module from cache
    return _; }
, connect: function(dsp) { // connects w/ initialized Espruino Grapics obj (after save()) 
    this.dsp=dsp;
    return this; }
};

if (Modules._eCache) { Modules._eCache(exports.mn, exports); } // for development in emulator only
