/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- uiRad ------------------------- 144 vars

ui radiobutton module provides radiobutton functionality to ui base module.

--- Enable ui with uiRad element functions:

Note: requires uiExt(ension) module (see uiExt.js module)

```
var ui = require("ui")       // load ui base module
      .add(require("iuExt")) // add uiExt module into base and remove from cache
      .add(require("uiRad")) // add uiRad module into base and remove from cache
      ;
```  

--- Create and add to ui three active(2), visible(1) =(3) radiobuttons at x/y (left/top:
5/80, 95/85 and 170/90) of different sizes/widths/diameters (37, 27 and 17) with size
related values ("L", "M" and "S") and labels ("Large", "Medium" and "Small"), with
group name "r1." (dot is part of group name). Medium sized radio button is set/checked
(first set/checked radiobutton wins... therefore small one is not set/checked even though
specified as such). Border colors (6, 3 and 5) and fill colors (2,1 and 2) are all
different (by choice). Unset/unchecked radiobuttons show ui display background color
(ui.bc). Label array definitions include fontVector (size), font color, x/y offset
from right/top (x2/y) corner of touch sensitive bounding box, and text.

Colors are bit-coded with 1-bit color-depth according ui.clrs=[...] setup.

```
ui.c(3,"rad","r1.a",  5, 80,  37,   0,   6,  2, "L",    cb,  [12, 7,   5,10, "Large"]);
ui.c(3,"rad","r1.b", 95, 85,  27,   1,   3,  1, "M",    0 ,  [12, 7,   5, 7, "Med"  ]);
ui.c(3,"rad","r1.c",170, 90,  17,   1,   5,  2, "S",    0 ,  [12, 7,   5, 2, "Small"]);
```

Callback cb is called on release (untouch) and provides
```id, v, ui, e, t``` as arguments:

```
args[] sym   description 
  0    id  : button id ("r1.a")
  1    v   : value object ("H" - can be any object - and undefined when unchecked)
  2    ui  : ui (THE ui object)
  3    e   : uiRad element (rad 'object' (runtime data structure) just released)
  4    t   : touch info x, y,... object to calc where within chk chk was 'hit' & released
        { x : x coordinate
        , y : y coordinate
        , t : touched (truey) / untouched (falsy)
        }
```

For detailed ui setup, including color depth and custom colors, connection to
display and touch screen or buttons, see ui module.

'constructor' is called by ```ui.c(f,"rad","radId01",...)``` with ```ui[args[1]+C](args);```
and 'creates' and registers rad ui element, conditionally displays it, and returns it.

rad ui element constructor arguments and runtime data structure are:

```
arg runtime 'object' instance of 'clazz' chk (chkbox)
a[]  e[]
 0   [0] flags focus(4), active(2), visible(1)
 .    .  0bxx1 visible  &1 visible 
 .    .  0bx1x active   &2 active / senses touch down (vs. read/display-only)
 .    .  0b1xx focus    &4 focus (by touchdown or drag in/over)  
 1   [1] clazz - "chk"
 2   [2] id - eg "c01", short, null or at least 3 chars, and ui globally unique
              Single letter ui element id's are 'reserved' (for keyboard(s)).
 3   [3] x  - x ((left ) of focus / touch bounding box)
 4   [4] y  - y ((top  ) of focus / touch bounding box)
 5       w  - width and hight (of focus / touch box,... 
     [5] x2 - x ((right) of focus / touch bounding box: x - w + 1)
 6       s  - initial checked(truey/1)/unchecked(falsy/0) state used w/ [9] value info
     [6] y2 - y ((bot  ) of focus / touch bounding box: y - h + 1)
 7   [7] bc - border color
 8   [8] fc - fill color
 9       v  - value (returned value when checked)
     [9] vi - value info array w/ state and value [truy/falsy=set/checked, value object]
         [0] - truy/falsy indicating state checked/unchecked state of radiobutton
         [1] - value object (returned when radiobutton is checked)
10  [10] cb - callback on untouch after touchdown w/ args (id, value, ui, e/btn, t/event)
11  [11] l  - label (info), array with:
      l[0]  fv - fontVector (size)
      l[1]  tc - (label) text color (index)
      l[2]  x  - x offset from focus box x ( bounding box left )
      l[3]  y  - y offset from focus box y ( bounding box top  )
      l[4]  tx - label text to display (using .drawString())
```

rad - radiobutton ui element methods (like prototype to mix into ui base):
 */
exports = // rad (radiobutton)
{ mn: "uiRad" // module 'clazz' name - globally unique (used to remove code from cache)
, radC:    function (f, c, i, x, y, d,     s,     bc, fc,  v,  cb,  l) { // constructor
    var gi=i.split(".")[0]+".", e=this.ff(this.es,function(e){ // find peer for group
      return ((e[1]==="rad")&&(e[10][0]===gi)); },3), g=(e)?e[10]:[gi,cb,null];
// ----------> rad e[0  1  2  3  4  5(x2)  6(y2)   7   8   9   10  11] - similar to args
    g.push(       e=[f, c, i, x, y, x+d-1, y+d-1, bc, fc,[s,v], g,  l]); // - runtime obj
    if (cb&&!g[1]) { g[1]=cb; } if (e[9][0]=s&&!g[2]) { g[2]=e; } return this.rdc(e); }
, radG: function(e) {
    var w=e[10]; return ((w=w[2])&&w[9][0]) ? w[1] : undefined; }
, radU: function(_,e,t,v,p) {
    var g=e[10], n=(v===undefined)?e:_.ff(g,function(o){ return (o[9][1]===v); },3), o;
    if (n && (n!==(o=g[2])) && (!o || (g===o[10]))) {
      if (o) { _.radu(_,o,o[9][0]=0); } _.radu(_,g[2]=n,n[9][0]=1);
      if (p) { if (g[1]&&g[1](n[2], n[9][1], _, n, t)) { _.d(n); } } } }
, radu: function(_,e,s) {
    _.radDs(_,e[3],e[4],Math.round((e[5]-e[3]-1)/2),e[8],s); }
, radD: function(f,c,i,x,y,x2,y2,bc,fc,vi,g,l) {
    var _=this, w2=Math.round((x2-x-1)/2),
    s=vi[0]; _.clr(bc).dsp.fillPoly(_.cvs(x+w2,y+w2,w2-2));
    if ((fc!==bc) || !s) { _.radDs(_,x,y,w2,fc,s); }
    if (l) { _.ld(x2,y,l); } }
, radDs: function(_,x,y,w2,fc,s) {
    _.clr((s)?fc:_.bc).dsp.fillPoly(_.cvs(x+w2,y+w2,w2-5)); }
, rad: function(_,e,t) {
    if ((e===_.ef) && !t.t) { _.radU(_,e,t,undefined,1); } }
};
