/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- ui_m ------------------------- 20190919


*** modular and extensible (graphical) ui framework - minimal version.


The framework consists of a bases module which provides the infrastructure
for individual out of box and custom ui elements and utility extensions.

Keeping separate versions not worth the 19 variables saved...
ui_ - regular - and ui_m - minimal versions main difference is foot print
and that with minimal foot print most ui elements support preferred 
callback ```on untouch``` only.

For further details see ui_ - regular version of ui (ui base module).

--- ui - base code into which elements and custom code are mixed in:

```

 */
exports = // ui base / 'DOM'/ui e(lement) data & code holder, singleton, for mixins)
{ dsp: null // display (Espruino Graphics object)
, mn: "ui"  // reserved, used temporary
, bc: 0     // dsp background color >=0: r*g*b* bit coded; <0: looked up [...,[R,G,B],...]
, tc: 7     // touch / focus color; >=0: r*g*b* bit coded; <0: looked up [...,[R,G,B],...]
, di: false // display instantly on create/change (not defer until after save())
, es: []    // ui elements
, ef: null  // (primary ui) element (in visual) focus (by touch down / hovering)
, lf: null  // (primary ui element) last (in visual) focus (by touch down / not hovering)
, af: null  // alternate (ui element in non-visual) focus (by drag over / hovering)
, it: false // (display) is touched
, td: false // is touch down (event)
, tt: 0     // touch down time
, lx: 0     // last x (touched)
, ly: 0     // last y (touched)
, dx: -1    // (last) x (touch) down
, dy: -1    // (last) y (touch) down
, te: {t:0} // (last) touch event
, clrs:     // default - bit coded color of 1 bit color-depth - 2^1^3(rgb) = 8 colors
    [function(c,i){ var v=(i)?c^7:c; return [v>>2&1,v>>1&1,v&1]; }.bind(exports)
    ]
, evt: function(x,y,_) { // common handling of touch / untouch / tap event
    _=_||this; var a=_.es, i=-1, m=a.length, e=_.ef, t={t:undefined!==x,f:0}; _.te=t;
    if (_.td=!_.it && t.t) { _.dx=x; _.dy=y; _.tt=getTime(); } // touch down
    if (_.it=t.t) { _.lx=t.x=x; _.ly=t.y=y; // --- touching ---
      if (e) { if (_.iib(e,x,y)) { return _[e[1]](_,e,t); } // xy event/move in focus elt
        _.blr(e) } // just moved out
      while (++i<m) { if (((e=a[i])[0]&2) && _.iib(e,x,y)) { // loop through active elts
        if (!(e[0]&4) && (_.td || (t.t && (e===_.lf)))) { _.foc(e); } // adj focus
        _[e[1]](_,e,t); return; } } if (_.td) { _.lf=null; } // handle hit / miss
    } else { if (e) { t.x=_.lx; t.y=_.ly; // --- untouch ---
      if (!_[e[1]](_,e,t)) { _.blr(e); } _.lf=null; _.dx=_.dy=-1; } } }
, iib: function(e,x,y) { // returns true if x/y is in bounding box of elt, else false
    return ((x>=e[3])&&(y>=e[4])&&(x<=e[5])&&(y<=e[6])); }
, foc: function(e,_) { // focus elt (unfocus elt in focus); visual: draw rectangle in...
    _=_||this; if (e!==_.ef) { if (_.ef) { _.blr(_.ef); } // ...touch color 'around' elt
    _.clr(_.tc).dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]|=4; _.ef=_.lf=e; }
    return _; }
, blr: function(e,_) { // blur elt (all if e falsy); visual: draw rectangle in display...
    _=_||this; _.clr(_.bc); (e?[e]:_.es).forEach(function(e){ // ...bg color around
        if (e[0]&4) { _.dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]&=11; } }); _.ef=null;
    return _; }
, rdc: function(e) { // register ui element, draw/display it conditionally, and return it
    this.es.push(e); if (((e[0]&=7)&1) && this.di) { this.d(e); }
    return e; }
, w: function(x,y,x2,y2,c) { // wipe rectangle (fill w/ c(olor) or display backgnd color)
    this.clr((c===undefined)?this.bc:c); this.dsp.fillRect(x,y,x2,y2); return this; }
, c: function() { // c(reate) ui element like DOM.createElement(), where arguments[1]...
    // ...defines ui element clazz and points to ui element specific constructor method
    var a=arguments, c=this[a[1]+"C"]; if (c) {
    return c.apply(this,a); } }  // undefined ui classes returns undefined as ui element
, d: function(e) { // display / redraw ui element e, redraw all if e falsy
    ((e)?[e]:this.es).forEach(function(e){ if (e[0]&1) this[e[1]+"D"].apply(this,e); },this);
    return this; }
, clr: function(c,i) { // set color according this.clrs[0](c,i) (customizable) function...
    this.dsp.setColor.apply(this.dsp,this.clrs[0](c,i)); // ...w/ color and i-nverse flag
    return this; }
, fnt: function(s) { // set displays's font vector (font size)
    this.dsp.setFontVector(s);
    return this; }
, ld: function(x,y,l) { // draw label (x, y, label[fv, tc, x, y, txt])
    this.fnt(l[0]).clr(l[1]).dsp.drawString(l[4],x+l[2],y+l[3]); }
, add: function(emon, k) { // add (mixin) non-overwriting properties of ui element by... 
    // eltModuleOrName for modularization and extensibility. Remove module from cache...
    var p, // ...when not k(eep) by element module name - if defined - to save space
    m = ("string" === typeof emon) ? require(emon) : emon; for (p in m) { if ((p!=="mn")
    &&!this[p]) { this[p] = m[p]; } } // merge / mixin properties 'new' (vars, functions)
    if (!k) { Modules.removeCached(m.mn); } // delete module from cache when not k(eep)
    return this; }
, connect: function(dsp) { // connects w/ initialized Espruino Grapics obj (after save()) 
    this.dsp=dsp;
    return this; }
};
