// ui0.js - ui

// 20190919 - allObjects - variables:  /  / 

// Modular and extensible UI - with base, ui elements -
// 2.4" 240x320 262K Color TFT LCD ILI9341 on SPI1 with 
// Resistive Touch Screen Controller XPT2046 on SPI2
// using out of box ILI9341 and ADS7843 modules
//
// Wiring:
//  2.4"
//  DSP                  PICO [<---USB
//   1   VCC      red     3   3.3            shared
//   2   GND      blk     1   GND            shared
//   3   CS       blu     7   B6   CS        LCD
//   4   RST      brn     9   A8   reset RST LCD
//   5   D/C      grn     8   B7   D/C       LCD
//   6   MOSI     ylw     6   B5   SPI1 MOSI LCD
//   7   SCK      wht     4   B3   SPI1 SCK  LCD
//   8   LED      org    18   A5   LED       LCD
//   9   MISO     grn     5   B4   SPI1 MISO LCD
//  10   T_CLK    wht    23   B13  SPI2 SCK  Touch
//  11   T_CS     ylw    22   B10  CS        Touch
//  12   T_MOSI   grn    25   B15  SPI2 MOSI Touch
//  13   T_MISO   blu    24   B14  SPI2 MISO Touch
//  14   T_IRQ    blk    21   B1   IRQ       Touch

// ----- pull in dispaly and touchscreen
var dW = 240, dH = 320, dsp, dMod = require("ILI9341"),
    touch, tMod = require("XPT2046");

// ----- pull in ui base and desired ui elements
var ui = require("ui")
      .add(require("uiBtn"))
      .add(require("uiSli"))
      .add(require("uiChk"))
      .add(require("uiInp"))
      .add(require("uiExt"))
      .add(require("uiRad"))
      ;

// ----- customize ui colors
ui.clrs = // imply color depth and table with converters and define user colors
  [function(c,i) { // color info to [r,g,b] converter
       var v, s=isNaN(c); if (s || c<0) { // c<0 (looked up [R,G,B]) or c=[R,G,B]
       return (((s) ? c : this.clrs[-c]).map(function(v){ return (i) ? v^255 : v/256; }));
     } else { // c>=0 bit coded rgb color (default 1 bit color depth 2^1^3=8 colors)
       v = (i) ? c^7 : c; return [v>>2&1,v>>1&1,v&1]; } }.bind(ui)     
  ,[216,216,216] // user color -1: light grey // keyboard bottom right special keys
  ,[192,192,192] // user color -2: medium light grey
  ,[128,128,128] // user color -3: medium gray
  ,[192,128,192] // user color -4: light purple
  ,[0  ,160, 64] // user color -5: kind of a 'darker' green
  ];
ui.bc = 0;     // ui (screen) background color
ui.di = false; // do not display instantly (on element creation), but deferred

// ----- define generic ui element callback - specific to this development context
var cb =  function(id,v) {
  if (v===undefined) { console.log(id); } else { console.log(id + ": " + v);  }
};


// ----- define UI (in level 0, so it does not use memory for source)

// clz:inp id active x   y   w   h   bc  fc     cb, vi, l (value info, label array obj)
//                                                   fv tc    x  y v(alue)  ml  tp  fm
ui.c("inp","i1" ,1, 45,  7,160, 25,   3,  7,    0 , [13, 0,   5, 4,"Jj"   , 16,  0,  0]
                                                  , [13, 7,-198, 4,"Field"]);

// clz:btn id active x   y   w   h   bc  fc     cb, l (label array obj)
//                                                   fv tc    x  y label text
ui.c("btn","b1" ,1,  5, 40, 65, 35,   4,  4,    cb, [15, 7,  13, 9,"RED"  ]);
ui.c("btn","b2" ,1, 75, 40, 65, 35,   5,  6,    cb, [15, 0,   9, 9,"Y-w-B"]);

ui.c("chk","c1" ,1,145, 47, 25,  0,  -5,  7,    cb, [12, 7,   2, 5,"OK"   ]);
ui.c("chk","c2" ,1,195, 47, 25,  1,   1,  1,    cb, [12, 7,   2, 5,"Up"   ]);

ui.c("rad","r1a",1,  5, 80, 35,  0,"L" ,  1,      6,  2,     cb, [12, 7,   5,10,"Large"]);
ui.c("rad","r1b",1, 95, 85, 25,  1,"M" ,  1,      3,  1,     0 , [12, 7,   5, 7,"Med"  ]);
ui.c("rad","r1c",1,170, 90, 15,  1,"S" ,  1,      5,  2,     0 , [12, 7,   5, 2,"Small"]);
ui.c("rad","r2a",1,  5,120, 20,  0,"L" ,  2,      1,  2,     cb, [12, 7,   5, 3,"Land" ]);
ui.c("rad","r2b",1, 95,120, 20,  0,"W" ,  2,     -3, -3,     0 , [12, 7,   5, 3,"Water"]);
ui.c("rad","r2c",1,170,120, 20,  1,"A" ,  2,      1,  1,     0 , [12, 7,   5, 3,"Air"  ]);

ui.c("sli","s1" ,1,  2,145,238, 25,  4, -5, cb, [[ 0, 0,  0, 0,   0,-10, 10]]);  
ui.c("sli","s2" ,1, 30,175,180, 25, -4, -4, cb, [[12, 6, 70, 6,  33,  0, 100, function(v){
                 var s=Math.round(v*10)+"%";
                 return ((v<1)?"0":"")+s.substr(0,s.length-2)+"."+s.substr(-2); } ]
                                                ,[12, 7,-25, 6, "%:0"]
                                                ,[12, 7,182, 6, "100"]
                                                ]);
ui.c("btn","b3" ,1,  3,287, 30, 30,               4,  4,     cb);
ui.c("btn","b4" ,1, 37,287, 30, 30,               2,  2,     cb);
ui.c("btn","b5" ,1, 71,287, 30, 30,               1,  1,     cb);
ui.c("btn","b6" ,1,105,287, 30, 30,               7,  7,     cb);
ui.c("btn","b7" ,1,139,287, 30, 30,               6,  6,     cb);
ui.c("btn","b8" ,1,173,287, 30, 30,               5,  5,     cb);
ui.c("btn","b9" ,1,207,287, 30, 30,               3,  3,     cb);


// get the keyboard and connect it to entry field - part of the UI definition
// {"uiKbd3x10B"}.connect(              ui,x,  y, w, h,bc,fc,sc,mx,my,fs,tc,tx,ty,cb) {
var kbd = require("uiKbd3x10B").connect(ui,0,215,24,24, 7, 7,-2, 1, 1,10 ,0, 2, 4,function(c){
   var s = ui.g("i1");
   if (c.length === 1) {
     ui.u("i1",s + "" + c);
   } else if (c === "bs") {
     if (s.length > 0) { ui.u("i1",s.substr(0,s.length - 1)); } 
   }
});

// var s2=ui.e("s2"); // get and hold on to slider w/ id "s2" for direct handling later on


// ----- run UI
function onInit() {
  SPI1.setup({sck:B3, miso:B4, mosi:B5, baud: 1000000});
                   // spi, dc, cs, rst, callback
  dsp = dMod.connect(SPI1, B7, B6, A8, function() {
      dsp.clear(); A5.set(); // display clear and back light on
      // connect ui to dsp and display it
      ui.connect(dsp)
        .w(0,0,dW-1,dH-1) // wipe (clear/fill) screen with screen background color
        .d() // display all elements
        .di=true; // set display immediately (display element immediately after creation)
      // setup touchscreen and connect it to ui
      SPI2.setup({sck:B13, miso:B14, mosi:B15, baud: 2000000});
      touch = tMod.connect(
          // spi, cs, irq, callback, calc
          SPI2, B10, B1, function(x,y) { ui.evt(x,y);
          }, function(yr, xr, d) {
            return [ Math.round(xr / -121.44          + 259.70685111989) 
                   , Math.round(yr /   88.90357142857 + -19.78130398103)
                   ];
      });
  });
}

onInit();
